/****************************************************************************************/
/* ESP32 DT-9501 Bluetooth Low Energy client                                            */
/* Version: 1.0                                                                         */
/* Date: 18.04.2023                                                                     */
/* Compiled with Arduino 1.8.16                                                         */
/* Arduino ESP32 Pico Kit                                                               */
/* Processor: ESP32                                                                     */
/* IMPORTANT!!! Partition scheme must be set to "No ATA (LargeAPP)                      */
/*                                                                                      */
/* PIN asignments:                                                                      */
/* GPIO 21 = Output = Red LED aktive High   - Bluetooth not paired                      */
/* GPIO 22 = Output = Green LED active high - Bluetooth paired                          */
/*                                                                                      */
/* Protocol explaination:                                                               */
/* Every time 15 bytes are received                                                     */
/* First 4 byte = Header always 0xD5, 0xF0, 0x00, 0x0A                                  */
/* 5th Byte = Date of device: Last 2 digits of year of device                           */
/* 6th Byte = Date of device: Month of device                                           */
/* 7th Byte = Date of device: Date of device                                            */
/* 8th byte = Time of device: Hour of device                                            */
/* 9th byte = Time of device: Minute of device                                          */
/* 10th byte = LSB: Digits 3 and 4 of device value                                      */
/* 11th byte = MSB: Digits 1 and 2 of device value                                      */
/* 12th byte = Always 0x00                                                              */
/* 13th byte = Position of digit point                                                  */
/* 14th byte = Measurement mode (see code for more details)                             */
/* 15th byte = Always 0x0D                                                              */
/*                                                                                      */
/* Bluetooth Low Energy example                                                         */
/* https://randomnerdtutorials.com/esp32-ble-server-client/                             */
/* https://www.mathaelectronics.com/esp32-ds18b20-tutorial-ds18b20-                     */
/* temperature-sensor-with-esp32/                                                       */
/****************************************************************************************/


/****************************************************************************************/
/* Libraries                                                                            */
/****************************************************************************************/
#include "BLEDevice.h"                            /* Include Bluetooth Low Enery (BLE) library */
#include "WiFi.h"                                 /* Include Wifi library */
#include "ESPAsyncWebServer.h"                    /* Include Async Web Server library */


/****************************************************************************************/
/* Definitions                                                                          */
/****************************************************************************************/
#define STATUS_LED_RED      21                    /* Pin 21 = Red LED   - Bluetooth not paired */ 
#define STATUS_LED_GREEN    22                    /* Pin 22 = Green LED - Bluetooth paired */
/* If this is uncommented here, debug info on processing the bluetooth connection and extraction of data being displayed! */
#define _DEBUG_


/****************************************************************************************/
/* Constants                                                                            */
/****************************************************************************************/
const char* SW_VERSION = "1.0";                   /* Version number */
const byte DATE_STRING_SIZE = 11;                 /* Reserve 11 byte for date string */
const byte TIME_STRING_SIZE = 6;                  /* Reserve 6 byte for time string  */
const byte MEASUREMENT_STRING_SIZE = 16;          /* Reserve 16 byte for measurement string */
/* The remote service of DT-9510 that needs to be used */
static BLEUUID serviceUUID("0000fff0-0000-1000-8000-00805F9B34FB");
/* The characteristic of the remote service of the DT-9510 that needs to be used */
static BLEUUID    charUUID("0000fff2-0000-1000-8000-00805F9B34FB");


/****************************************************************************************/
/* Variables                                                                            */
/****************************************************************************************/
char*    RECEIVE_STRING;                          /* Char array for reception of data from DT-9501 */
byte     RECEIVE_LENGTH;                          /* Length the received data string from DT-9501 --> Should be 15 */
boolean  RECEIVED_FLAG;                           /* Received flag: True = Data received, False = No data received */
String   DATE_STRING;                             /* Date string extracted from DT-9501 device */ 
String   TIME_STRING;                             /* Time string extracted from DT-9501 device */
String   MEASUREMENT_STRING;                      /* Measurement string extracted from DT-9501 device */
static boolean DO_CONNECT_FLAG = false;           /* True = The scan for devices was performed and the desired device was found */
static boolean CONNECTED_FLAG  = false;           /* True = The device is connected */
static boolean DO_SCAN_FLAG    = false;           /* True = A scan should be performed */
/* Variable for remote characteristic */
static BLERemoteCharacteristic* pRemoteCharacteristic;
/* Define advertised device as DT-9501 */
static BLEAdvertisedDevice* DT_9501;


/****************************************************************************************/
/* Definitions and variables for Wifi                                                   */
/****************************************************************************************/
const char* WIFI_SSID = "DT-9501";                /* Enter Wifi SSID here */
const char* WIFI_PASSWORD = "12345678";           /* Enter Wifi Password here */
IPAddress LOCAL_IP (192,168,1,4);                 /* Enter IP Address details of hotspot here */
IPAddress LOCAL_GW (192,168,1,1);                 /* Enter Gateway IP Address details of hotspot here */
IPAddress LOCAL_SN (255,255,255,0);               /* Enter Subnet IP Address details of hotspot here */


/****************************************************************************************/
/* Initalization webserver                                                              */
/****************************************************************************************/
AsyncWebServer server(80);                        /* Initialize Async Web server */
AsyncEventSource events("/events");               /* Initialize Async Web server event update */


/****************************************************************************************/
/****************************************************************************************/
/* Code for website                                                                     */
/****************************************************************************************/
/****************************************************************************************/
const char INDEX_HTML[] PROGMEM = R"rawliteral(
  <!DOCTYPE HTML><html>
  <head>
    <title>DT-9510 Web Server</title>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="stylesheet" href="https://use.fontawesome.com/releases/v5.7.2/css/all.css" integrity="sha384-fnmOCqbTlWIlj8LyTjo7mOUStjsKC4pOpQbqyi7RrhN7udi9RwhKkMHpvLbHG9Sr" crossorigin="anonymous">
    <link rel="icon" href="data:,">
    <style>
      html {font-family: Arial; display: inline-block; text-align: center;}
      p {  font-size: 1.2rem;}
      body {  margin: 0;}
      .topnav { overflow: hidden; background-color: #0066CC; color: white; font-size: 1.7rem; }
      .content { padding: 20px; }
      .card { background-color: white; box-shadow: 2px 2px 12px 1px rgba(140,140,140,.5); }
      .cards { max-width: 700px; margin: 0 auto; display: grid; grid-gap: 2rem; grid-template-columns: repeat(auto-fit, minmax(300px, 1fr)); }
      .readinga { font-size: 2.8rem; }
      .readingb { font-size: 2.5rem; }
      .card.measurement { color: #5f81a9; }
      .card.date { color: #696969; }
    </style>
  </head>
  <body>
    <div class="topnav">
      <h3>DT-9510 WEB SERVER</h3>
    </div>
    <div class="content">
      <div class="cards">
        <div class="card measurement">
          <h4> VALUE</h4><p><span class="readinga"><span id="meas">%MEASUREMENT%</p>
        </div>
        <div class="card date">
          <h4> TIME</h4><p><span class="readingb"><span id="tim">%TIME%</p>
          <h4> DATE</h4><p><span class="readingb"><span id="dat">%DATE%</p>
        </div>
      </div>
    </div>
    <script>
      if (!!window.EventSource) 
      {
        var source = new EventSource('/events');
 
        source.addEventListener('open', function(e) 
        {
          console.log("Events Connected");
        }, false);
        source.addEventListener('error', function(e) 
        {
          if (e.target.readyState != EventSource.OPEN) 
          {
            console.log("Events Disconnected");
          }
        }, false);
        source.addEventListener('message', function(e) 
        {
          console.log("message", e.data);
        }, false);
        source.addEventListener('measurement', function(e) 
        {
          console.log("measurement", e.data);
          document.getElementById("meas").innerHTML = e.data;
        }, false);
        source.addEventListener('date', function(e) 
        {
          console.log("date", e.data);
          document.getElementById("dat").innerHTML = e.data;
        }, false);
        source.addEventListener('time', function(e) 
        {
          console.log("time", e.data);
          document.getElementById("tim").innerHTML = e.data;
        }, false);
      }
    </script>
  </body>
</html>)rawliteral";


/****************************************************************************************/
/****************************************************************************************/
/* Bluetooth callcack functions                                                         */
/****************************************************************************************/
/****************************************************************************************/

/****************************************************************************************/
/* Callback function to search for BLE servers and find the first one that              */
/* advertises the service that was defined to search for                                */
/****************************************************************************************/
class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks 
{
   /* Called for each advertising BLE server */
  void onResult(BLEAdvertisedDevice advertisedDevice) 
  {
    Serial.print("  The following BLE advertised device found: ");
    Serial.println(advertisedDevice.toString().c_str());
    /* After a device is found, it needs to check if it contains the service as it was specified */
    if ((advertisedDevice.haveServiceUUID() && advertisedDevice.isAdvertisingService(serviceUUID))) 
    {
      BLEDevice::getScan()->stop();               /* Stop scanning device */
      DT_9501 = new BLEAdvertisedDevice(advertisedDevice);
      DO_CONNECT_FLAG = true;                     /* Set start connect flag true */
      DO_SCAN_FLAG = true;                        /* Set do scan flag true */
    }
  }
};


/****************************************************************************************/
/* Callback function for notification when new data from decive is available            */
/* The data is inside the array pData with the length length                            */
/****************************************************************************************/
static void notifyCallback (BLERemoteCharacteristic* pBLERemoteCharacteristic, uint8_t* pData, size_t length, bool isNotify) 
{ 
  RECEIVED_FLAG = false;                          /* Set received flag false */
  RECEIVE_LENGTH = length;                        /* Copy length of pData into the global variable */
  if (RECEIVE_LENGTH >= 15)                       /* 15 bytes or more received ? */
  {
    RECEIVE_STRING = (char*)pData;                /* Copy array into a global char array */
    RECEIVED_FLAG = true;                         /* Set received flag true */
  }
}


/****************************************************************************************/
/* Callback function if client gets connected or disconnected                           */
/* The data is inside the array pData with the length length                            */
/****************************************************************************************/
class MyClientCallback : public BLEClientCallbacks 
{
  void onConnect(BLEClient* pclient)              /* Connected ? --> Do nothing */
  {
  }

  void onDisconnect(BLEClient* pclient)           /* Disonnected ? */
  {
    CONNECTED_FLAG = false;                       /* Set connected flag false */
    Serial.println("DT-9501 got disconnected !!!");
  }
};


/****************************************************************************************/
/****************************************************************************************/
/* Subroutines                                                                          */
/****************************************************************************************/
/****************************************************************************************/

/****************************************************************************************/
/* Subroutine to replace placeholder with variable values for display of website        */
/****************************************************************************************/
String SUB_WEB_PROCESSOR (const String& REPLACE_TARGET)
{
  if(REPLACE_TARGET == "MEASUREMENT")             /* Keyword measurement detected ? */
  {
    return (MEASUREMENT_STRING);                  /* Replace variable with measurement string */
  }
  if(REPLACE_TARGET == "DATE")                    /* Keyword date detected ? */
  {
    return (DATE_STRING);                         /* Replace variable with date string */
  }
  if(REPLACE_TARGET == "TIME")                    /* Keyword time detected ? */
  {
    return (TIME_STRING);                         /* Replace variable with time string */
  }
}


/****************************************************************************************/
/* Subroutine to connect to DT9501                                                      */
/* Return true if connection is established                                             */
/****************************************************************************************/
bool SUB_CONNECT_DT9501() 
{
  #ifdef _DEBUG_  
    Serial.println ();
    Serial.print("Found device DT-9501! Starting a connection to BLE address: ");
    Serial.println(DT_9501->getAddress().toString().c_str());
  # endif   
  /* Step A: Create a client */
  BLEClient*  pClient  = BLEDevice::createClient();
  #ifdef _DEBUG_ 
    Serial.println("  Client was created successfully!");
  # endif
  /* Step B: Attach callbacks to client */
  pClient->setClientCallbacks(new MyClientCallback());
  #ifdef _DEBUG_ 
    Serial.println("  Callbacks attached successfully!");
  # endif  
  /* Step C: Connect to DT-9501 */
  pClient->connect(DT_9501);
  #ifdef _DEBUG_ 
    Serial.println("  Connected to DT-9501!");
  # endif  
  /* Step D: Obtain a reference to the service of the DT-9501 */
  BLERemoteService* pRemoteService = pClient->getService(serviceUUID);
  if (pRemoteService == nullptr)                  /* Service not found ? */
  {
    Serial.print ("ERROR: Failed to find service UUID ");
    Serial.print (serviceUUID.toString().c_str());
    Serial.println (" of DT-9501");
    pClient->disconnect();                        /* Disconnect from DT-9501 */
    return false;                                 /* Return false */
  }
  #ifdef _DEBUG_ 
    Serial.print ("  Found service UID ");
    Serial.print (serviceUUID.toString().c_str());
    Serial.println (" of DT-9501");
  # endif  
  /* Step E: Obtain a reference to the characteristic in the service of the DT-9501 */
  pRemoteCharacteristic = pRemoteService->getCharacteristic(charUUID);
  if (pRemoteCharacteristic == nullptr)           /* Characteristic not found ? */
  {
    Serial.print("ERROR: Failed to find characteristic UUID ");
    Serial.print(charUUID.toString().c_str());
    Serial.println (" of DT-9501");
    pClient->disconnect();                        /* Disconnect from DT-9501 */
    return false;                                 /* Return false */
  }
  #ifdef _DEBUG_ 
    Serial.print ("  Found characteristic UID ");
    Serial.print(charUUID.toString().c_str());
    Serial.println (" of DT-9501");
  # endif
  /* Step F: Read the value of the characteristic */
  if(pRemoteCharacteristic->canRead())            /* Value of characteristic can be read ? */
  {
    /* Read characteristic value */
    std::string value = pRemoteCharacteristic->readValue();
    #ifdef _DEBUG_ 
      Serial.print("  The characteristic value is: ");
      Serial.println(value.c_str());
    # endif
  }
  /* Step G: Enable notify */
  if(pRemoteCharacteristic->canNotify())
      pRemoteCharacteristic->registerForNotify(notifyCallback);
  #ifdef _DEBUG_
    Serial.println("  Notify is enabled!");
    Serial.println();
  # endif
  Serial.println ("Ready to receive data...");
  Serial.println ();
  CONNECTED_FLAG = true;                          /* Set connected flag true */
  return true;                                    /* Return false */
}


/****************************************************************************************/
/* Subroutine to decode received data and trigger events for display on Webpage         */
/****************************************************************************************/
void SUB_DECODE_MESSAGE ()
{
  boolean ERROR_FLAG;                             /* True = Error during receptiom, False = No error during reception */
  byte    CALC_BYTE;                              /* Byte for calculation */

  ERROR_FLAG = false;                             /* Set error flag false */
  DATE_STRING = "";                               /* Delete date string */
  TIME_STRING = "";                               /* Delete time string */
  MEASUREMENT_STRING = "";                        /* Delete measurement string */
  
  if (RECEIVE_LENGTH == 15)                       /* Right number of bytes received */
  {
    #ifdef _DEBUG_
      Serial.print ("Received correct ");
      Serial.print (RECEIVE_LENGTH);
      Serial.println (" bytes of data! ");    
    # endif 
    /* Incorrect start header not received ? */
    if ((RECEIVE_STRING[0] != 0xD5) || (RECEIVE_STRING[1] != 0xF0) || (RECEIVE_STRING[2] != 0x00) || (RECEIVE_STRING[3] != 0x0A))         
    {
      Serial.print ("ERROR: Wrong start header ");
      Serial.print("0x");
      Serial.print(RECEIVE_STRING[0] < 16 ? "0" : "");
      Serial.print (RECEIVE_STRING[0], HEX);
      Serial.print(" 0x");
      Serial.print(RECEIVE_STRING[1] < 16 ? "0" : "");
      Serial.print (RECEIVE_STRING [1], HEX);
      Serial.print(" 0x");        
      Serial.print(RECEIVE_STRING[2] < 16 ? "0" : "");
      Serial.print (RECEIVE_STRING [2], HEX);
      Serial.print(" 0x"); 
      Serial.print(RECEIVE_STRING[3] < 16 ? "0" : "");
      Serial.print (RECEIVE_STRING [3], HEX);
      Serial.println(" received!"); 
      ERROR_FLAG = true;                          /* Set error flag true */ 
    }   
      #ifdef _DEBUG_                        
        else                                      /* Correct start header received ? */
        {
          Serial.println ("Correct start header received!");
        }                
      # endif 
  }
  else                                            /* Incorrect number of bytes  received ? */
  {
    Serial.print ("ERROR: Received wrong ");
    Serial.print (RECEIVE_LENGTH);
    Serial.println (" bytes of data! ");
    Serial.println();
    ERROR_FLAG = true;                            /* Set error flag true */   
  }

  /* Correct end and in between header */
  if (ERROR_FLAG == false)                        /* No error --> Check next step */
  {
    /* Correct end header and in between header not received ? */
    if ((RECEIVE_STRING[14] != 0x0D)|| (RECEIVE_STRING[11] != 0x00))
    {
      ERROR_FLAG = true;                          /* Set error flag true */  
      Serial.print ("ERROR: Wrong end or in between header ");
      Serial.print(RECEIVE_STRING[11] < 16 ? "0" : "");
      Serial.print (RECEIVE_STRING [11], HEX);
      Serial.print(RECEIVE_STRING[14] < 16 ? "0" : "");
      Serial.print (RECEIVE_STRING [14], HEX);
      Serial.println(" received!");
    }                                         
      #ifdef _DEBUG_   
        else                                      /* Correct header received ? */
          {
            Serial.println ("Correct end and in between header received!");
          }     
      # endif 
  }  

  /* Cummulate data */
  if (ERROR_FLAG == false)                        /* No error --> Check next step */
  {
    /* Calculate date */
    if ((int)RECEIVE_STRING [6]< 10)
    { 
      DATE_STRING += 0;                    
    }
    DATE_STRING += (int)RECEIVE_STRING [6];
    DATE_STRING += ".";
    if (RECEIVE_STRING [5]< 10)
    { 
      DATE_STRING += 0;                   
    }
    DATE_STRING += (int)RECEIVE_STRING [5];
    DATE_STRING += ".20";
    if ((int)RECEIVE_STRING [4]< 10)
    { 
      DATE_STRING += 0;                     
    }
    DATE_STRING += (int)RECEIVE_STRING [4];
    Serial.print ("Received date stamp: ");
    Serial.println(DATE_STRING); 
      
    /* Calculate time */
    if ((int)RECEIVE_STRING [7]< 10)
    { 
      TIME_STRING += 0;                      
    }
    TIME_STRING += (int)RECEIVE_STRING [7];
    TIME_STRING += ":";
    if ((int)RECEIVE_STRING [8]< 10)
    { 
      TIME_STRING += 0;                     
    }
    TIME_STRING += (int)RECEIVE_STRING [8];
    Serial.print ("Received time stamp: ");
    Serial.println(TIME_STRING); 

    /* Calculate measurement value */
    #ifdef _DEBUG_  
      Serial.print ("Measurement bytes: 0x");       
      Serial.print(RECEIVE_STRING[10] < 16 ? "0" : "");
      Serial.print (RECEIVE_STRING [10], HEX);
      Serial.print(" 0x");        
      Serial.print(RECEIVE_STRING[9] < 16 ? "0" : "");
      Serial.print (RECEIVE_STRING [9], HEX);
      Serial.print (" received, dezimal point position: "); 
      Serial.println ((int)RECEIVE_STRING [12]);
    # endif             
    /* Get 1st digit */
    CALC_BYTE = RECEIVE_STRING [10] & 0xF0;
    CALC_BYTE = CALC_BYTE >> 4; 
    if ((int)RECEIVE_STRING [12] == 3)            /* Dezimalpoint at position 3?--> Display leading zero anyway */  
    {          
      MEASUREMENT_STRING += CALC_BYTE;
      MEASUREMENT_STRING += ".";
    }
    else                                          /* Decimalpoint not at position 3? */
    {
      if (CALC_BYTE != 0)                         /* Display digit if not zero? */
        MEASUREMENT_STRING += CALC_BYTE;
    }
    /* Get second digit */     
    CALC_BYTE = RECEIVE_STRING [10] & 0x0F;
    if ((int)RECEIVE_STRING [12] == 2)            /* Dezimalpoint at position 2\ --> Display leading zero anyway */  
    {          
      MEASUREMENT_STRING += CALC_BYTE;
      MEASUREMENT_STRING += ".";
    }
    else                                          /* Decimalpoint not at position 2? */
    {
      /* Display digit if not zero and decimal point not on 1 or 0? */
      if ((CALC_BYTE != 0) ||  (((int)RECEIVE_STRING [12] != 1) &&  ((int)RECEIVE_STRING [12] != 0)))    
        MEASUREMENT_STRING += CALC_BYTE;
    }
    /* Get third digit */   
    CALC_BYTE = RECEIVE_STRING [9] & 0xF0;
    CALC_BYTE = CALC_BYTE >> 4;             
    if ((int)RECEIVE_STRING [12] == 1)            /* Dezimalpoint at position 1? --> Display leading zero anyway */  
    {          
      MEASUREMENT_STRING += CALC_BYTE;
      MEASUREMENT_STRING += ".";
    }
    else                                          /* Decimalpoint not at position 2? */
    {
      /* Display digit if not zero and decimal point not on 1 or 0? */            
      if ((CALC_BYTE != 0) || (((int)RECEIVE_STRING [12] != 1) &&  ((int)RECEIVE_STRING [12] != 0)))            
        MEASUREMENT_STRING += CALC_BYTE;
    }
    /* Get forth digit */  
    CALC_BYTE = RECEIVE_STRING [9] & 0x0F;
    MEASUREMENT_STRING += CALC_BYTE; 
    /* Display according right mode */
    switch(RECEIVE_STRING [13])
    {
      case 1:
        #ifdef _DEBUG_ 
          Serial.println ("Mode 1: Infinite dose rate accumlation in µSv!");    
        # endif                     
        MEASUREMENT_STRING += " µSv";
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                              
        break;
      case 2:  
        #ifdef _DEBUG_             
          Serial.println ("Mode 2: Timed dose rate accumlation in µSv started!"); 
        # endif                 
        MEASUREMENT_STRING += " µSv";
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                                   
        break;             
      case 3:  
        #ifdef _DEBUG_             
          Serial.println ("Mode 3: Timed dose rate accumlation in µSv stopped!"); 
        # endif  
        MEASUREMENT_STRING += " µSv";
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                                 
        break;  
      case 4:  
        #ifdef _DEBUG_             
          Serial.println ("Mode 4: Current radiation dose in µSv/h!");
        # endif                                
        MEASUREMENT_STRING += " µSv/h";
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                   
        break;             
      case 5:  
        #ifdef _DEBUG_             
          Serial.println ("Mode 5: Current radiation dose in Rem/h!"); 
        # endif                  
        MEASUREMENT_STRING += " mR/h";
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                     
        break;   
      case 6: 
        #ifdef _DEBUG_ 
          Serial.println ("Mode 6: Pulse count per second in CPS!");  
        # endif                 
        MEASUREMENT_STRING += " CPS";
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                                 
        break;   
      case 7:
        #ifdef _DEBUG_ 
          Serial.println ("Mode 7: Pulse count per minute in CPM!");
        # endif                   
        MEASUREMENT_STRING += " CPM";
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                                 
        break;  
      case 8:  
        #ifdef _DEBUG_             
          Serial.println ("Mode 8: Pulse counting mode!"); 
        # endif                   
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                                
        break; 
      case 9:  
        #ifdef _DEBUG_             
          Serial.println ("Mode 9: Timed pulse counting mode started!");
        # endif                  
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                                 
        break;  
      case 10:  
        #ifdef _DEBUG_             
          Serial.println ("Mode 10: Timed pulse counting mode stopped!");
        # endif                   
        Serial.print ("Measurement result: "); 
        Serial.println (MEASUREMENT_STRING);                                                
        break;               
      default :  
        Serial.println ("ERROR: Unknown mode!"); 
        ERROR_FLAG = true;                        /* Set error flag true */                                     
        break;  
    }                                                                                   
  }
  Serial.println();

  /* Send events to the Web Server with the three strings */
  if (ERROR_FLAG == false)                        /* No error --> Check next step */
  {
    events.send("ping",NULL,millis());
    events.send(MEASUREMENT_STRING.c_str(),"measurement",millis());
    events.send(DATE_STRING.c_str(),"date",millis());
    events.send(TIME_STRING.c_str(),"time",millis());
  }
}


/****************************************************************************************/
/* Setup procedure                                                                      */
/****************************************************************************************/ 
void setup() 
{
  /* Reserve byte buffer for measurement string */
  MEASUREMENT_STRING.reserve (MEASUREMENT_STRING_SIZE);  
  DATE_STRING.reserve (DATE_STRING_SIZE);         /* Reserve byte buffer for date string */ 
  TIME_STRING.reserve (TIME_STRING_SIZE);         /* Reserve byte buffer for time string */
  MEASUREMENT_STRING = "----";                    /* Initialize measurement string */
  DATE_STRING = "--.--.----";                     /* Initialize date string */
  TIME_STRING = "--.--";                          /* Initialize time string */
    
  /* Configure outputs for LEDs */
  pinMode(STATUS_LED_RED, OUTPUT);                 /* Configure Status Led 1 as output */
  digitalWrite(STATUS_LED_RED, HIGH);              /* Switch Status Led 1 on */
  pinMode(STATUS_LED_GREEN, OUTPUT);               /* Configure Status Led 2 as output */
  digitalWrite(STATUS_LED_GREEN, LOW);             /* Switch Status Led 2 off */
  /* Initialize variables */
  RECEIVE_LENGTH = 0;                              /* Reset length of received message */
  RECEIVED_FLAG = false;                           /* Set received flag false */
  /* Initialize serial interface and display welcome message */
  Serial.begin(115200);
  /* Display welcome message with initialisation information */
  Serial.println(F("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"));
  Serial.println (F("*********************************************************"));
  Serial.println (F("* Bluetooth client for radiation meter DT-9501          *"));
  Serial.println (F("*                                                       *"));
  Serial.print   (F("* V"));
  Serial.print   (SW_VERSION);
  Serial.println (F("                                      @PT 2023    *"));   
  Serial.println (F("*********************************************************\n")); 

  /* Initialize the Wifi access point to access the web page */
  WiFi.softAP(WIFI_SSID, WIFI_PASSWORD);           /* Initialize the soft API */
  /* Configure the soft API */
  WiFi.softAPConfig( LOCAL_IP, LOCAL_GW, LOCAL_SN);
  delay(100);                                      /* Wait 100ms to complete setup */
  Serial.println ("Setting up WIFI access point...");
  Serial.print ("  Connect to WIFI access point ");
  Serial.print (WIFI_SSID);
  Serial.print (" and enter password ");
  Serial.print (WIFI_PASSWORD);
  Serial.print (" and then type IP adress '");
  Serial.print (LOCAL_IP);
  Serial.println ("' into the browser to display webpage!");
  /* Handle Web Server */
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request)
  {
    request->send_P(200, "text/html", INDEX_HTML, SUB_WEB_PROCESSOR);
  });
  /* Handle Web Server events */
  events.onConnect([](AsyncEventSourceClient *client)
  {
    if(client->lastId())
    {
      Serial.printf("Client reconnected! Last message ID that it got is: %u\n", client->lastId());
    }
    /* Step Send event with message "Hello!", id current millis and set reconnect delay to 1 second */
    client->send("Hello!", NULL, millis(), 10000);
  });
   
  server.addHandler(&events);                      /* Add server events */
  server.begin();                                  /* Start server */
   
  Serial.println("\nStarting to scan BLE devices...");
  BLEDevice::init("");                             /* Initialize BLE application */
  /* Retrieve a Scanner and set the callback to be informed when a new device is detected */
  /* Specify active scanning and start the scan to run for 5 seconds. */
  BLEScan* pBLEScan = BLEDevice::getScan();
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setInterval(1349);
  pBLEScan->setWindow(449);
  pBLEScan->setActiveScan(true);  
  pBLEScan->start(0, false);  
}


/****************************************************************************************/
/* Main Loop                                                                            */
/****************************************************************************************/
void loop() 
{
  /* Check if the DO_CONNECT_FLAG is true after the BLE Server is connected */
  if (DO_CONNECT_FLAG == true) 
  {
    if (SUB_CONNECT_DT9501())                      /* Connected ? */
    {
      digitalWrite(STATUS_LED_RED, LOW);           /* Switch Status Led 1 off */
      digitalWrite(STATUS_LED_GREEN, HIGH);        /* Switch Status Led 2 on */     
    } 
    else 
    {
      Serial.println("ERROR!!! No connection possible! Reseting the device in 5 seconds!");
      delay (5000);                                /* Wait for 5 seconds */
      ESP.restart();                               /* ESP32 software reset */
    }
    DO_CONNECT_FLAG = false;                       /* Set do connect flag false */
  }
  

  if (CONNECTED_FLAG)                              /* DT-9501 device connected ? */
  {
    if (RECEIVED_FLAG == true)                     /* Data have been received ? */
    {
      SUB_DECODE_MESSAGE ();                       /* Call subroutine to decode reveived value */
      RECEIVED_FLAG = false;                       /* Set received flag false */
    }
  }
  else if(DO_SCAN_FLAG)
  {
    digitalWrite(STATUS_LED_RED, HIGH);            /* Switch Status Led 1 on */
    digitalWrite(STATUS_LED_GREEN, LOW);           /* Switch Status Led 2 off */
    BLEDevice::getScan()->start(0);                /* Start scan after disconnect! */  
  }
}
